// 12LEDcircle.c
// written by dale wheat - 18 may 2011
// changed from ATtiny13A to ATtiny25 on 30 May 2016

#include <avr/io.h>
#include <avr/eeprom.h>

///////////////////////////////////////////////////////////////////////////////
// notes:
//	device = ATtiny25, internal RC oscillator 8 MHz / 8 (1.0 MHz)
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
//
// Charlieplexed LED array
//
// LEDs are arranged in a circle and numbered like a clock face
//
//
//          (12)
//     (11)      ( 1)
//
//  (10)            ( 2)
//
// ( 9)              ( 3)
//
//  ( 8)            ( 4)
//
//     ( 7)      ( 5)
//          ( 6)
//
// Single LED logic table:
//
// LED hi lo
//   1  1  2
//   2  2  1
//   3  2  3
//   4  3  2
//   5  3  1
//   6  1  3
//   7  1  4
//   8  4  1
//   9  4  2
//  10  2  4
//  11  4  3
//  12  3  4
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// macros
///////////////////////////////////////////////////////////////////////////////

#define nop() asm("nop")

#define sbi(port, bit) (port) |= (1 << (bit))
#define cbi(port, bit) (port) &= ~(1 << (bit))

///////////////////////////////////////////////////////////////////////////////
// mode definitions
///////////////////////////////////////////////////////////////////////////////

typedef enum {

	MODE_DEMO,
	MODE_RING_FADE,
	MODE_TICK_1,
	MODE_TICK_2,
	MODE_TICK_3,
	MODE_TICK_4,
	MODE_RANDOM_FAST,
	MODE_RANDOM_1,
	MODE_RANDOM_SPARKLE,
	MODE_FILL,
	MODE_GHOST_CIRCLE,
	MODE_MAX, // always the last enumerated mode value
} MODE;

volatile MODE mode __attribute__ ((section (".noinit")));

uint8_t eeprom_do_not_use EEMEM; // bad luck - do not use location 0
uint8_t eeprom_mode EEMEM; // mode storage location

///////////////////////////////////////////////////////////////////////////////
// init() - initialize everything
// note:  this "function" is in section .init3 so it is executed before main()
///////////////////////////////////////////////////////////////////////////////

void init(void) __attribute__ ((naked, section(".init3")));
void init(void) {

	// turn off unused peripherals to save power

	PRR = 1<<PRTIM1 | 1<<PRTIM0 | 1<<PRUSI | 1<<PRADC;
	ACSR = 1<<ACD; // disable analog comparator
	DIDR0 = 1<<ADC3D | 1<<ADC2D | 1<<ADC1D | 1<<ADC0D | 1<<AIN1D | 1<<AIN0D; // disable all digital inputs

	// EEPROM address register is always pointed to the "mode memory" address, 1

	asm("ldi r24, 1"); // EEPROM address register always points to mode memory location: 1
	asm("out 0x1e, r24");

	if(bit_is_set(MCUSR, PORF)) {
		sbi(EECR, EERE); // strobe EEPROM read enable signal
		mode = EEDR; // read mode memory location from EEPROM
		if(mode > MODE_MAX) {
			mode = MODE_DEMO; // demonstration mode
		}
	} else if(bit_is_set(MCUSR, EXTRF)) {
		mode++; // advance mode
		if(mode > MODE_MAX) {
			mode = MODE_DEMO; // reset mode
		}
		//eeprom_write_byte(&eeprom_mode, mode); // memorize
		EEDR = mode; // move mode to EEPROM data register, address register is already set up
		asm("out 0x1c, r1"); // write a zero to the EEPROM control register
		sbi(EECR, EEMPE); // enable the "master program" bit
		sbi(EECR, EEPE); // write the data to the EEPROM
	}

	MCUSR = 0; // reset bits

	// initialize input & output port

	// PORTB
	//			SOIC8
	//	name	DIP8	QFN20	other functions				usage
	//	----	----	-----	----------------			-----
	//	PB0		5		11		MOSI/AIN0/OC0A/PCINT0		LED matrix node 1
	//	PB1		6		12		MISO/AIN1/OC0B/INT0/PCINT1	LED matrix node 2
	//	PB2		7		14		SCK/ADC1/T0/PCINT2			LED matrix node 3
	//	PB3		2		2		PCINT3/CLKI/ADC3			LED matrix node 4
	//	PB4		3		5		PCINT4/ADC2					n/c
	//	PB5		1		1		PCINT5/-RESET/ADC0/dW		-RESET button

	PORTB = 0<<PORTB5 | 0<<PORTB4 | 0<<PORTB3 | 0<<PORTB2 | 0<<PORTB1 | 0<<PORTB0;
	DDRB = 0<<DDB5 | 0<<DDB4 | 0<<DDB3 | 0<<DDB2 | 0<<DDB1 | 0<<DDB0;
}

///////////////////////////////////////////////////////////////////////////////
// data tables
///////////////////////////////////////////////////////////////////////////////

struct {
	uint8_t ddr:4; // data direction register
	uint8_t pin:4; // port pin data register
} table[] = {
	{ 0b0011, 0b0001 }, // 1
	{ 0b0011, 0b0010 }, // 2
	{ 0b0110, 0b0010 }, // 3
	{ 0b0110, 0b0100 }, // 4
	{ 0b0101, 0b0100 }, // 5
	{ 0b0101, 0b0001 }, // 6
	{ 0b1001, 0b0001 }, // 7
	{ 0b1001, 0b1000 }, // 8
	{ 0b1010, 0b1000 }, // 9
	{ 0b1010, 0b0010 }, // 10
	{ 0b1100, 0b1000 }, // 11
	{ 0b1100, 0b0100 }, // 12
	{ 0b0000, 0b0000 }, // 0 (all LEDs off)
};

uint8_t ghost[] = {

	0b0010, // 1
	0b0001, // 2
	0b0100, // 3
	0b0010, // 4
	0b0001, // 5
	0b0100, // 6
	0b1000, // 7
	0b0001, // 8
	0b0010, // 9
	0b1000, // 10
	0b0100, // 11
	0b1000, // 12
	0b0000, // all LEDs off
};

#define ALL_LEDS_OFF 12
#define off() PORTB = 0x00 // all LEDs off

///////////////////////////////////////////////////////////////////////////////
// led() - illuminate zero or one LEDs (0-11 = LEDs, 12 = all LEDs off)
///////////////////////////////////////////////////////////////////////////////

void led(uint8_t n, uint8_t duration) {

	volatile uint8_t i;

	off();
	DDRB = table[n].ddr;
	PORTB = table[n].pin;

	i = duration;

	while(i) i--; // short delay
	off(); // anti-ghosting
}

///////////////////////////////////////////////////////////////////////////////
// led_pwm() - illuminate zero or one LEDs with variable brightness
///////////////////////////////////////////////////////////////////////////////

void led_pwm(uint8_t n, uint8_t brightness) {

	uint8_t i;

	for(i = 0; i < 255; i++) {

		if(i < brightness) {

			DDRB = table[n].ddr; // output direction bits
			PORTB = table[n].pin; // set first node pin high

		} else {

			DDRB = table[ALL_LEDS_OFF].ddr; // no outputs
			PORTB = table[ALL_LEDS_OFF].pin; // all LEDs off
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// ring() - variable brightness ring function; emulates all LEDs on
///////////////////////////////////////////////////////////////////////////////

void ring(uint8_t brightness) {

	uint8_t i;

	for(i = 0; i < 255; i++) {

		if(i < brightness) {

			DDRB = 0b1111; // all outputs

			PORTB = 0b0011; // 3+6+7+10
			PORTB = 0b0101; // 1+4+7+12
			PORTB = 0b0110; // 2+5+10+12
			PORTB = 0b1001; // 1+6+9+11
			PORTB = 0b1010; // 2+3+8+11
			PORTB = 0b1100; // 4+5+8+9

			off(); // all LEDs off

		} else {

			// darkness;  also, take the same amount of time as above

			DDRB = 0;
			PORTB = 0;
			PORTB = 0;
			PORTB = 0;
			PORTB = 0;
			PORTB = 0;
			PORTB = 0;
			PORTB = 0;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////
// ring_fade() - fade the ring from off to on then back to off again
///////////////////////////////////////////////////////////////////////////////

void ring_fade(void) {

	uint8_t i;

	for(i = 0; i < 255; i++) ring(i); // fade up
	for(i = 255; i > 0; i--) ring(i); // fade down
}

///////////////////////////////////////////////////////////////////////////////
// spin() - illuminate all LEDs in sequence, with anti-aliasing
///////////////////////////////////////////////////////////////////////////////

//void spin(void) {
//
//	uint8_t i, i1, i2, j;
//
//	for(i = 0; i < 12; i++) {
//
//		i1 = i;
//		i2 = i + 1;
//		if(i2 == 12) i2 = 0;
//
//		for(j = 0; j < 10; j++) {
//			led(i1, 128);
//			led(i2, 128);
//		}
//		for(j = 0; j < 20; j++) {
//			led(i1, 255);
//		}
//	}
//
//	off(); // all LEDs off
//}

///////////////////////////////////////////////////////////////////////////////
// tick() - light 1, 2, 3 or 4 LEDs equidistantly around the perimeter, twice
///////////////////////////////////////////////////////////////////////////////

uint8_t tick_steps[] = {12, 6, 4, 3};
uint8_t tick_delay[] = {60, 30, 20, 15};

void tick(uint8_t n) {

	uint8_t cycle, i;
	uint8_t delay;

	for(cycle = 0; cycle < n<<1; cycle++) {
		for(i = 0; i < tick_steps[n - 1]; i++) {
			for(delay = tick_delay[n - 1]; delay > 0; delay--) {
				led(i, 255); // primary LED
				if(n == 2) {
					led(i + 6, 255); // secondary LED
				} else if(n == 3) {
					led(i + 4, 255); // tertiary LEDs
					led(i + 8, 255);
				} else if(n == 4) {
					led(i + 3, 255); // quaternary LEDs
					led(i + 6, 255);
					led(i + 9, 255);
				}
			}
		}
	}

	off(); // all LEDs off
}

///////////////////////////////////////////////////////////////////////////////
// pseudorandom number generator
///////////////////////////////////////////////////////////////////////////////

uint16_t prand(void) {

    static uint16_t prand_value = 0xDA1E; // randomly seeded ;)
    
    prand_value = (prand_value >> 1) ^ (-(prand_value & 1) & 0xd001);

    return prand_value;
}

///////////////////////////////////////////////////////////////////////////////
// random_fast() - blink all LEDs both randomly & quickly
///////////////////////////////////////////////////////////////////////////////

void random_fast(void) {

	uint16_t n = 1500; // about 5 seconds

	while(n) {
		led_pwm(prand() % 12, 255); // pick an LED, any LED
		n--;
	}

	off(); // all LEDs off
}

///////////////////////////////////////////////////////////////////////////////
// random_one() - fade a single LED at random
///////////////////////////////////////////////////////////////////////////////

void random_one(void) {

	uint8_t i = 18 /* about 10 sceonds */, n, brightness;

	while(i) {
		
		n = prand() % 12; // pick an LED
		for(brightness = 255; brightness > 0; brightness -= 3) {
			led_pwm(n, brightness);
		}
		i--;
	}
}

///////////////////////////////////////////////////////////////////////////////
// random_sparkle() - display 1-4 LEDs for random time - looks sparkly
///////////////////////////////////////////////////////////////////////////////

void random_sparkle(void) {

	uint8_t i = 150; // about 5 seconds
	volatile uint16_t time;

	while(i) {
		DDRB = prand();
		PORTB = prand();
		time = prand() & 0x0FFF;
		while(time) time--; // random delay
		i--;
	}

	off(); // all LEDs off
}

///////////////////////////////////////////////////////////////////////////////
// scan() - scan LED buffer
///////////////////////////////////////////////////////////////////////////////

uint8_t buffer[12] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

void scan(void) {

	uint8_t i;

	for(i = 0; i < 12; i++) {
		off(); // anti-ghosting
		if(buffer[i]) {
			DDRB = table[i].ddr;
			PORTB = table[i].pin;
		} else {
			DDRB = table[ALL_LEDS_OFF].ddr; // off
			PORTB = table[ALL_LEDS_OFF].ddr; // off
		}
	}

	off(); // all LEDs off
}

///////////////////////////////////////////////////////////////////////////////
// fill() - fill then unfill circle
///////////////////////////////////////////////////////////////////////////////

void fill(void) {

	uint8_t i;
	uint16_t j;

	for(i = 11; i < 255; i--) {
		buffer[i] = 255; // fill
		for(j = 0; j < 1000; j++) scan();
	}
	for(i = 11; i < 255; i--) {
		buffer[i] = 0; // unfill
		for(j = 0; j < 1000; j++) scan();
	}
}

///////////////////////////////////////////////////////////////////////////////
// crescent() - ratcheting crescent shape
///////////////////////////////////////////////////////////////////////////////

void crescent(void) {

	// *** TO DO ***
}

///////////////////////////////////////////////////////////////////////////////
// ghost_led() - barely illuminate a single LED
///////////////////////////////////////////////////////////////////////////////

void ghost_led(uint8_t i) {

	DDRB = ghost[i];
	PORTB = table[i].pin;
}

///////////////////////////////////////////////////////////////////////////////
// ghost_circle() - spin dim LED around circle
///////////////////////////////////////////////////////////////////////////////

void ghost_circle(void) {

	uint8_t i;
	volatile uint16_t j;

	for(i = 0; i < 12; i++) {
		ghost_led(i);
		for(j = 0; j < 10000; j++);
	}
}

///////////////////////////////////////////////////////////////////////////////
// demo() - show off all the (viually interesting) modes
///////////////////////////////////////////////////////////////////////////////

void demo(void) {

	// fade the entire ring up & down 3 times

	ring_fade();
	ring_fade();
	ring_fade();

//	// spin an LED around the circle 3 times
//
//	spin();
//	spin();
//	spin();

	// walk one or more LEDs around the circumference

	tick(1);
	tick(2);
	tick(3);
	tick(4);

	random_fast(); // blink LEDs furiously at random

	random_one(); // fade random LEDs one at a time

	random_sparkle(); // random sparkle (1-4 LEDs)

	// fill and unfill circle

	fill();
	fill();
	fill();
	fill();

	crescent(); // ratcheting crescent

	// ghost circle

	ghost_circle();
	ghost_circle();
	ghost_circle();
	ghost_circle();
}

///////////////////////////////////////////////////////////////////////////////
// main() - main program function
///////////////////////////////////////////////////////////////////////////////

void main(void) {

	//mode = MODE_FILL; // ** debug *** test fill() function

	// the main program loop

	while(1) {

		//led(3, 255); continue; // *** debug *** test led() function
		//led_pwm(0, 1); continue; // *** debug *** test led_pwm() function
		//debug_led(); // *** debug *** test led() function
		//ring(255); // *** debug *** test ring() function
		//ring_fade(); // *** debug *** test ring_fade() function
		//leds(1, 1, 2, 20); // *** debug *** test leds() function
		//spin(); // *** debug *** test spin() function
		//tick(1); //continue; // *** debug *** test tick() function
		//tick(2); //continue;
		//tick(3); //continue;
		//tick(4); continue;
		//random_fast(); // *** debug *** test random_fast() function
		//random_one(); // *** debug *** test random_one() function
		//random_sparkle(); // *** debug *** test random_sparkle() function
		//fill(); // *** debug *** test fill() function
		//ghost_circle(); continue; // *** debug *** test ghost_circle() function

		switch(mode) {

		case MODE_DEMO: // demonstration mode
			demo();
			break;

		case MODE_RING_FADE: // fade the entire ring up and down
			ring_fade();
			break;

		case MODE_TICK_1:
			tick(1);
			break;

		case MODE_TICK_2:
			tick(2);
			break;

		case MODE_TICK_3:
			tick(3);
			break;

		case MODE_TICK_4:
			tick(4);
			break;

		case MODE_RANDOM_FAST:
			random_fast();
			break;

		case MODE_RANDOM_1:
			random_one();
			break;

		case MODE_RANDOM_SPARKLE:
			random_sparkle();
			break;

		case MODE_FILL:
			fill();
			break;

		case MODE_GHOST_CIRCLE:
			ghost_circle();
			break;

		case MODE_MAX: // off
			DDRB = 0x00; // all LEDs off
			PORTB = 0x00;
			while(1) {
				// deepest sleep mode
				MCUCR = 1<<PUD | 1<<SE | 1<<SM1 | 0<<SM0 | 0<<ISC01 | 0<<ISC00; // select "power down" mode
				asm("sleep"); // go to sleep to save power
			}
			break;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////

// [end-of-file]

